/*
 * alestis_event_classifier
 *
 * Beacons event classifier for Alestis.
 *
 * Copyright(c) 2019 BeOneOff Digital Solutions S.L.
 * All rights reserved.
 *
 * Unauthorized copying of this file, via any medium is strictly prohibited.
 * Proprietary and confidential.
 *
 * Authors:
 * Alberto Castillo <alberto.castillo@beoneoff.com>
 */

const moment = require('moment');
const BeaconModel = require('../models/beacon.model');

/**
 * Get the current beacon or create a new beacon
 * @param {BeaconModel} beaconArray
 * @param {JSObject} data -> From Broker*
 * @returns {BeaconModel}
 */
function currentBeaconOrNewBeacon(beaconArray, data) {
  let found = false;
  let currentBeacon;
  let i = 0;
  while (!found && i < beaconArray.length) {
    if (data.ID === beaconArray[i].id) {
      found = true;
      currentBeacon = beaconArray[i];
      currentBeacon.addRSSI(data.RSSI);
      currentBeacon.updateLastSeen(data.TIME);
      return currentBeacon;
    }
    i += 1;
  }
  if (!found) {
    const newBeacon = new BeaconModel(data.ID, data.TIME);
    beaconArray.push(newBeacon);
    currentBeacon = newBeacon;
    currentBeacon.addRSSI(data.RSSI);
    return currentBeacon;
  }
  return null;
}

/**
 * Push & Pop the beacon in Array based on the State.
 * @param {Array} oldArray
 * @param {Array} newArray
 * @param {BeaconModel} beacon
 */
function moveBeacon(oldArray, newArray, beacon) {
  const beaconIndex = oldArray.indexOf(beacon);
  oldArray.splice(beaconIndex, 1);
  newArray.push(beacon);
}

/**
 * Get the difference in seconds between NOW and the Beacon's Last Seen
 * @param {BeaconModel} beacon
 * @returns {Number} in seconds
 */
function differenceBetweenNowLastSeen(beacon) {
  const nowMoment = moment();
  const lastSeen = moment(beacon.getLastSeen()).utc();
  const difference = nowMoment.diff(lastSeen, 'seconds');
  return difference;
}

/**
 * Change beacon's state based on a condition
 * @param {BeaconModel} beacon
 * @param {Number} leaveToEnterDelay
 * @param {Number} enterToAliveDelay
 * @param {Array} enterBeacons
 * @param {Array} aliveBeacons
 * @param {Array} leaveBeacons
 */
function stateTransaction(
  beacon,
  leaveToEnterDelay,
  enterToAliveDelay,
  enterBeacons,
  aliveBeacons,
  leaveBeacons,
) {
  const difference = differenceBetweenNowLastSeen(beacon);
  switch (beacon.state) {
    // LEAVE => ENTER
    // TODO RSSI FILTER
    case 'LEAVE':
      if (difference <= leaveToEnterDelay) {
        moveBeacon(leaveBeacons, enterBeacons, beacon);
        beacon.setState('ENTER');
      }
      break;
    case 'ALIVE':
      break;
    // ENTER => ALIVE
    // TODO RSSI FILTER
    case 'ENTER':
      if (difference < enterToAliveDelay) {
        moveBeacon(enterBeacons, aliveBeacons, beacon);
        beacon.setState('ALIVE');
      } else {
        moveBeacon(enterBeacons, leaveBeacons, beacon);
        beacon.setState('LEAVE');
      }
      break;
    default:
      break;
  }
}

/**
 * Is the beacon leaving?
 * @param {Number} rssi
 * @param {Number} thresholdMax
 * @param {Number} thresholdMin
 * @returns {Boolean}
 */
function isLeaving(rssi, thresholdMax, thresholdMin) {
  let result = false;
  if (rssi >= thresholdMax && rssi <= thresholdMin) {
    result = true;
  }
  return result;
}

/**
 * Is the beacon alive?
 * YES => Change the state to LEAVE
 * NO => Break the execution
 * @param {Array} aliveBeacons - Beacons which are alive
 * @param {Array} leaveBeacons - Beacons which are leave
 * @param {Number} thresholdMax - RSSI ThresholdMax
 * @param {Number} thresholdMin - RSSI ThresholMin
 */
function isBeaconKeepAlive(aliveBeacons, leaveBeacons, thresholdMax, thresholdMin, delay) {
  if (aliveBeacons.length > 0) {
    aliveBeacons.forEach((beacon) => {
      const last = beacon.kalmanRSSI.length - 1;
      const lastRSSI = beacon.kalmanRSSI[last];
      const nextToLastRSSI = beacon.kalmanRSSI[last - 1];
      const isLeavingLast = isLeaving(lastRSSI, thresholdMax, thresholdMin);
      const isLeavingNextToLast = isLeaving(nextToLastRSSI, thresholdMax, thresholdMin);
      if (isLeavingLast && isLeavingNextToLast) {
        const difference = differenceBetweenNowLastSeen(beacon);
        if (difference > delay) {
          moveBeacon(aliveBeacons, leaveBeacons, beacon);
          beacon.setState('LEAVE');
        }
      }
    });
  }
}

module.exports = {
  currentBeaconOrNewBeacon,
  stateTransaction,
  isBeaconKeepAlive,
};
